"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.createBrowser = void 0;
const path_1 = require("path");
const resolveExports = require("resolve.exports");
const resolveFile = require("resolve");
const jsdom_1 = require("jsdom");
const context_require_1 = require("context-require");
const noopModule = (0, path_1.join)(__dirname, "../noop.js");
const exportsMainFile = `__package_exports__`;
const resolveExportsOptions = {
    unsafe: true,
    browser: true,
    conditions: ["default", "require", "browser"],
};
/**
 * Creates a custom Module object which runs all required scripts
 * in a new jsdom instance.
 */
function createBrowser({ dir, html, extensions, beforeParse, ...jsdomOptions }) {
    const browser = new jsdom_1.JSDOM("", {
        runScripts: "dangerously",
        pretendToBeVisual: true,
        ...jsdomOptions,
    });
    const { window } = browser;
    const resolveExtensions = extensions
        ? [
            ...new Set([
                ...Object.keys(extensions),
                ...Object.keys(require.extensions).reverse(),
            ]),
        ]
        : Object.keys(require.extensions).reverse();
    // Pass through istanbul coverage.
    if ("__coverage__" in globalThis) {
        window.__coverage__ = globalThis.__coverage__;
    }
    // Expose some globals commonly shimmed by bundlers.
    window.global = window;
    window.Buffer = Buffer;
    window.process = { ...process, browser: true };
    window.setImmediate || (window.setImmediate = (() => {
        const msg = `${Math.random()}`;
        let queue = [];
        let offset = 0;
        window.addEventListener("message", (ev) => {
            if (ev.data === msg) {
                const cbs = queue;
                offset += cbs.length;
                queue = [];
                for (const cb of cbs) {
                    if (cb) {
                        cb();
                    }
                }
            }
        });
        window.clearImmediate = function clearImmediate(id) {
            const index = id - offset;
            if (index >= 0) {
                queue[index] = undefined;
            }
        };
        return function setImmediate(cb, ...args) {
            const index = queue.push(args.length ? () => cb(...args) : cb) - 1;
            if (!index) {
                window.postMessage(msg, "*");
            }
            return index + offset;
        };
    })());
    browser.require = (0, context_require_1.default)({
        dir,
        context: browser,
        extensions,
        resolve(basedir, req, { filename }) {
            return resolveFile.sync(req, {
                basedir,
                filename,
                pathFilter,
                packageFilter,
                extensions: resolveExtensions,
            });
        },
    });
    browser.yield = () => new Promise((resolve) => window.setImmediate(resolve));
    browser.act = async (fn) => {
        let errors;
        let errorStage = 0;
        window.addEventListener("error", handleError);
        window.addEventListener("unhandledrejection", handleError);
        try {
            const result = await fn?.();
            await browser.yield();
            return result;
        }
        catch (err) {
            trackError(err);
        }
        finally {
            window.removeEventListener("error", handleError);
            window.removeEventListener("unhandledrejection", handleError);
            switch (errorStage) {
                case 0:
                    break;
                case 1:
                    throw errors;
                default:
                    throw new window.AggregateError(errors);
            }
        }
        function handleError(ev) {
            if (!ev.defaultPrevented) {
                let error = "error" in ev ? ev.error : ev.reason || ev;
                if ("detail" in error)
                    error = error.detail;
                ev.preventDefault();
            }
        }
        function trackError(err) {
            switch (errorStage) {
                case 0:
                    errors = err;
                    errorStage = 1;
                    break;
                case 1:
                    if (err !== errors) {
                        errors = new Set([errors, err]);
                        errorStage = 2;
                    }
                    break;
                default:
                    errors.add(err);
                    break;
            }
        }
    };
    if (beforeParse) {
        beforeParse(window, browser);
    }
    window.document.open();
    window.document.write(html || "<!DOCTYPE html><html><head></head><body></body></html>");
    return browser;
    function pathFilter(pkg, _file, relativePath) {
        if (pkg.exports) {
            return resolveExports.exports(pkg, relativePath === exportsMainFile ? "." : relativePath, resolveExportsOptions)?.[0];
        }
        else if (pkg.browser) {
            const pkgBrowser = pkg.browser;
            let requestedFile = relativePath === exportsMainFile ? "." : relativePath;
            if (typeof pkgBrowser === "string") {
                switch (requestedFile) {
                    case ".":
                    case "./":
                        return pkgBrowser;
                    default:
                        return requestedFile;
                }
            }
            let replacement = pkgBrowser[requestedFile];
            if (replacement !== undefined) {
                return replacementOrNoop(replacement);
            }
            if (requestedFile === ".") {
                // Check `.` and `./`
                requestedFile = "./";
                replacement = pkgBrowser[requestedFile];
                if (replacement !== undefined) {
                    return replacementOrNoop(replacement);
                }
            }
            else if (requestedFile[0] !== ".") {
                requestedFile = `./${requestedFile}`;
                replacement = pkgBrowser[requestedFile];
                if (replacement !== undefined) {
                    return replacementOrNoop(replacement);
                }
            }
            const isFolder = requestedFile[requestedFile.length - 1] === "/";
            if (isFolder) {
                // If we're definitely matching a folder we'll try adding `index` to the
                // end first.
                requestedFile += "index";
                replacement = pkgBrowser[requestedFile];
                if (replacement !== undefined) {
                    return replacementOrNoop(replacement);
                }
            }
            for (const ext of resolveExtensions) {
                replacement = pkgBrowser[requestedFile + ext];
                if (replacement !== undefined) {
                    return replacementOrNoop(replacement);
                }
            }
            if (!isFolder) {
                // If we're not matching a folder we'll try adding `/index` to the end.
                requestedFile += "/index";
                replacement = pkgBrowser[requestedFile];
                if (replacement !== undefined) {
                    return replacementOrNoop(replacement);
                }
                for (const ext of resolveExtensions) {
                    replacement = pkgBrowser[requestedFile + ext];
                    if (replacement !== undefined) {
                        return replacementOrNoop(replacement);
                    }
                }
            }
        }
        return relativePath;
    }
}
exports.createBrowser = createBrowser;
function packageFilter(pkg) {
    if (pkg.exports || pkg.browser) {
        // defers to the "exports" field.
        pkg.main = exportsMainFile;
    }
    return pkg;
}
function replacementOrNoop(id) {
    return id === false ? noopModule : id;
}
